@@ -8,6 +8,7 @@ from ticdat.utils import ForeignKey, ForeignKeyMapping, TypeDictionary, verify,
 from ticdat.utils import lupish, deep_freeze, containerish, FrozenDict, safe_apply, stringish
 import ticdat.pandatio as pandatio
 from itertools import count
+import math
 try:
     from pandas import isnull
     import numpy
@@ -22,20 +23,29 @@ except:
 
 pd, DataFrame = utils.pd, utils.DataFrame # if pandas not installed will be falsey
 
-def _faster_df_apply(df, func, trip_wire_check=None):
-    cols = list(df.columns)
-    data, index = [], []
-    for row in df.itertuples(index=True):
-        row_dict = {f:v for f,v in zip(cols, row[1:])}
-        data.append(func(row_dict))
-        index.append(row[0])
-        if trip_wire_check:
-            new_func = trip_wire_check(data[-1])
-            if new_func:
-                func = new_func
-                trip_wire_check = None
-    # will default to float for empty Series, like original pandas
-    return pd.Series(data, index=index, **({"dtype": numpy.float64} if not data else {}))
+def _is_last_rows_all_nan(df, last_rows):
+    assert last_rows > 0
+    # quick last row check to make faster
+    all_nan_row =  lambda row: all(map(pd.isnull, row.values()))
+    return utils.faster_df_apply(df.tail(1), all_nan_row).all() and \
+           utils.faster_df_apply(df.tail(last_rows), all_nan_row,
+                            trip_wire_check=lambda x: all_nan_row if x else lambda row: False).all()
+
+def _find_number_all_nan_last_rows(df, min_rtn, max_rtn):
+    assert max_rtn >= min_rtn
+    if min_rtn == max_rtn:
+        assert min_rtn == 0 or _is_last_rows_all_nan(df, min_rtn)
+        return min_rtn
+    med_rtn =  math.ceil((min_rtn+max_rtn)/2)
+    if _is_last_rows_all_nan(df, med_rtn):
+        return _find_number_all_nan_last_rows(df, med_rtn, max_rtn)
+    return _find_number_all_nan_last_rows(df, min_rtn, med_rtn-1)
+
+def remove_trailing_all_nan(df):
+    all_nan_last_rows = _find_number_all_nan_last_rows(df, 0, len(df))
+    if all_nan_last_rows:
+        return df.head(len(df)-all_nan_last_rows).copy(deep=True)
+    return df
 
 class PanDatFactory(object):
     """
@@ -94,7 +104,10 @@ class PanDatFactory(object):
                 "default_values" : self.default_values,
                 "data_types" : self.data_types,
                 "parameters": self.parameters,
-                "infinity_io_flag": self.infinity_io_flag}
+                "infinity_io_flag": self.infinity_io_flag,
+                "xlsx_trailing_empty_rows": self.xlsx_trailing_empty_rows,
+                "duplicates_ticdat_init": self.duplicates_ticdat_init,
+                "tooltips": utils.make_tooltips_dict_json_friendly(self.tooltips)}
     @staticmethod
     def create_from_full_schema(full_schema):
         """
@@ -109,7 +122,8 @@ class PanDatFactory(object):
         """
         old_schema = {"tables_fields", "foreign_keys", "default_values", "data_types"}
         verify(dictish(full_schema) and set(full_schema).issuperset(old_schema) and set(full_schema) in
-               utils.all_subsets(old_schema.union({"parameters", "infinity_io_flag"})),
+               utils.all_subsets(old_schema.union({"parameters", "infinity_io_flag", "xlsx_trailing_empty_rows",
+                                                   "duplicates_ticdat_init", "tooltips"})),
                "full_schema should be the result of calling schema(True) for some PanDatFactory")
         fks = full_schema["foreign_keys"]
         verify( (not fks) or (lupish(fks) and all(lupish(_) and len(_) >= 3 for _ in fks)),
@@ -144,25 +158,111 @@ class PanDatFactory(object):
                 rtn.add_parameter(p, *((df,) + tuple(dt)), enforce_type_rules=True)
         if "infinity_io_flag" in full_schema:
             rtn.set_infinity_io_flag(full_schema["infinity_io_flag"])
+        if "xlsx_trailing_empty_rows" in full_schema:
+            rtn.set_xlsx_trailing_empty_rows(full_schema["xlsx_trailing_empty_rows"])
+        if "duplicates_ticdat_init" in full_schema:
+            rtn.set_duplicates_ticdat_init(full_schema["duplicates_ticdat_init"])
+        if "tooltips" in full_schema:
+            for t, tip_dict in full_schema["tooltips"].items():
+                for f, tip in tip_dict.items():
+                    rtn.set_tooltip(t, f, tip)
         return rtn
-    def clone(self, table_restrictions=None):
+    def clone(self, table_restrictions=None, clone_factory=None):
         """
 
         clones the PanDatFactory
+
         :param table_restrictions : if None, then argument is ignored. Otherwise, a container listing the
                                     tables to keep in the clone. Tables outside table_restrictions are removed from
                                     the clone.
-        :return: a clone of the PanDatFactory
+
+        :param clone_factory : optional. Defaults to PanDatFactory. Can also be TicDatFactory.  Can also be a function,
+                               in which case it should behave similarly to create_from_full_schema.
+                               If clone_factory=TicDatFactory, the row predicates that use predicate_kwargs_maker
+                               won't be copied over.
+
+        :return: a clone of the PanDatFactory. Returned object will be based on clone_factory, if provided.
+
+        Note - If you want to remove tables via a clone, then call like this
+               pdf_new = pdf.clone(table_restrictions=set(pdf.all_tables).difference(tables_to_remove))
+               Other schema editing operations are available with clone_add_a_table, clone_add_a_column,
+               clone_remove_a_column and clone_rename_a_column.
         """
-        full_schema = utils.clone_a_anchillary_info_schema(self.schema(include_ancillary_info=True), table_restrictions)
-        rtn = PanDatFactory.create_from_full_schema(full_schema)
+        clone_factory = clone_factory or PanDatFactory
+        from ticdat import TicDatFactory
+        no_copy_predicate_kwargs_maker = clone_factory == TicDatFactory
+        if hasattr(clone_factory, "create_from_full_schema"):
+            clone_factory = clone_factory.create_from_full_schema
+        full_schema = utils.clone_a_anchillary_info_schema(self.schema(include_ancillary_info=True),
+                                                           table_restrictions)
+        rtn = clone_factory(full_schema)
         for tbl, row_predicates in self._data_row_predicates.items():
             if table_restrictions is None or tbl in table_restrictions:
                 for pn, rpi in row_predicates.items():
-                    rtn.add_data_row_predicate(tbl, predicate=rpi.predicate, predicate_name=pn,
-                                               predicate_kwargs_maker=rpi.predicate_kwargs_maker,
-                                               predicate_failure_response=rpi.predicate_failure_response)
+                    if not (rpi.predicate_kwargs_maker and no_copy_predicate_kwargs_maker):
+                        rtn.add_data_row_predicate(tbl, predicate=rpi.predicate, predicate_name=pn,
+                                                   predicate_kwargs_maker=rpi.predicate_kwargs_maker,
+                                                   predicate_failure_response=rpi.predicate_failure_response)
         return rtn
+    def clone_add_a_table(self, table, pk_fields, df_fields):
+        '''
+
+        add a table to the PanDatFactory
+
+        :param table: table not in the schema
+
+        :param pk_fields: container of the primary key fields
+
+        :param df_fields: container of the data fields
+
+        :return: a clone of the PanDatFactory, with the new table added
+        '''
+        return utils.clone_add_a_table(self, table, pk_fields, df_fields)
+    def clone_add_a_column(self, table, field, field_type, field_position="append"):
+        '''
+
+        add a column to the PanDatFactory
+
+        :param table: table in the schema
+
+        :param field: name of the new field to be added
+
+        :param field_type: either "primary key" or "data"
+
+        :param field_position: integer between 0 and the length of self.primary_key_fields[table] (if "primary key")
+                               or self.data_fields[table] (if "data"), inclsuive.
+                               Alternately, can be "append", which will just insert the column at the end of the
+                               appropriate list.
+
+        :return: a clone of the PanDatFactory, with field inserted into location field_position for field_type
+        '''
+        return utils.clone_add_a_column(self, table, field, field_type, field_position)
+    def clone_remove_a_column(self, table, field):
+        '''
+        remove a column from the PanDatFactory
+
+        :param table: table in the schema
+
+        :param field: name of the field to be removed
+
+        :return: a clone of the PanDatFactory, with field removed
+        '''
+        return utils.clone_remove_a_column(self, table, field)
+    def clone_rename_a_column(self, table, field, new_field):
+        '''
+        rename a column in the PanDatFactory
+
+        :param table: table in the schema
+
+        :param field: name of the field to be removed
+
+        :param new_field: new name for the field
+
+        :return: a clone of the PanDatFactory, with field renamed to new_field. Data types, default values and
+                 foreign keys will reflect the new field name, but row predicates will be copied over as-is (and thus you will need
+                 to re-create them as needed).
+        '''
+        return utils.clone_rename_a_column(self, table, field, new_field)
     @property
     def default_values(self):
         return deep_freeze(self._default_values)
@@ -171,9 +271,68 @@ class PanDatFactory(object):
         return utils.FrozenDict({t : utils.FrozenDict({k :v for k,v in vd.items()})
                                 for t,vd in self._data_types.items()})
     @property
+    def tooltips(self):
+        return utils.FrozenDict(self._tooltips)
+
+    def set_tooltip(self, table, field, tooltip):
+        """
+        Set the tooltip for a table, or for a (table, field) pair.
+
+        :param table: a table in the schema
+
+        :param field: an empty string (if you want to set the tooltip for a table)
+                      or a field for this table
+
+        :param tooltip: an empty string (if you want to delete a previously set tooltip) or the
+                        tooltip you want to set
+
+        :return:
+
+        After calling this function, the tooltips property for this PanDatFactory will be appropriately adjusted.
+        """
+        utils.set_tooltip(self, table, field, tooltip, self._tooltips)
+
+    @property
     def parameters(self):
         return FrozenDict(self._parameters)
     @property
+    def duplicates_ticdat_init(self):
+        """
+        see __doc__ for set_duplicates_ticdat_init
+        """
+        return self._duplicates_ticdat_init[0]
+    def set_duplicates_ticdat_init(self, value):
+        """
+        Set the duplicates_ticdat_init for the PanDatFactory. Choices are:
+        --> 'assert' : an assert is raised if duplicate rows are passed to TicDat.__init__
+        --> 'warn'   : emit a warning if duplicate rows are passed to TicDat.__init__
+        --> 'ignore' : don't do anything if duplicate rows are passed to TicDat.__init__
+        This is only relevant when using copy_to_tic_dat
+        :param value: either 'assert', 'warn' or 'ignore'
+        :return:
+        """
+        verify(value in ["assert", "warn", "ignore"], f"bad value {value}")
+        self._duplicates_ticdat_init[0] = value
+    @property
+    def xlsx_trailing_empty_rows(self):
+        """
+        see __doc__ for set_xlsx_trailing_empty_rows
+        """
+        return self._xlsx_trailing_empty_rows[0]
+    def set_xlsx_trailing_empty_rows(self, value):
+        """
+        Set the xlsx_trailing_empty_rows for the PanDatFactory. Choices are:
+        --> 'prune' : (the default) when reading an xlsx/xlsm file, look for trailing all pd.isnull rows in each table,
+                      and prune them
+        --> 'ignore': retain such rows
+        With the move to openpyxl for xlsx/xlsm file reading, its more likely that Excel users accidentally creating
+        trailing all none rows.
+        :param value: either 'prune' or 'ignore'
+        :return:
+        """
+        verify(value in ["prune", "ignore"], f"bad value {value}")
+        self._xlsx_trailing_empty_rows[0] = value
+    @property
     def infinity_io_flag(self):
         """
         see __doc__ for set_infinity_io_flag
@@ -181,7 +340,7 @@ class PanDatFactory(object):
         return  self._infinity_io_flag[0]
     def set_infinity_io_flag(self, value):
         """
-        Set the infinity_io_flag for the TicDatFactory.
+        Set the infinity_io_flag for the PanDatFactory.
         'N/A' (the default) is recognized as a flag to disable infinity I/O buffering.
 
         If numeric, when writing data to the file system (or a database), float("inf") will be replaced by the
@@ -248,16 +407,26 @@ class PanDatFactory(object):
         :param json_read: special 'None'->None override needed for pandas json reader
         '''
         assert push_parameters_to_be_valid or not json_read, "json_read should always push_parameters_to_be_valid"
-        apply = _faster_df_apply
+        apply = utils.faster_df_apply
         for t in set(self.all_tables).difference(["parameters"]): # parameters table is handled differently
             df = getattr(dat, t)
-            for f in self.primary_key_fields.get(t, ()) + self.data_fields.get(t, ()):
-                if utils.numericish(self.infinity_io_flag):
+            all_fields = tuple(self.primary_key_fields.get(t, ()) + self.data_fields.get(t, ()))
+            if utils.numericish(self.infinity_io_flag):
+                fields_w_issues = set()
+                def find_fields_w_issues(row):
+                    for f in all_fields:
+                        if utils.numericish(row[f]) and abs(row[f]) >= self.infinity_io_flag:
+                            fields_w_issues.add(f)
+                apply(df, find_fields_w_issues)
+                for f in fields_w_issues:
                     fixme = apply(df, lambda row: utils.numericish(row[f]) and row[f] >= self.infinity_io_flag)
-                    df.loc[fixme, f] = float("inf")
+                    if fixme.any():
+                        df.loc[fixme, f] = float("inf")
                     fixme = apply(df, lambda row: utils.numericish(row[f]) and row[f] <= -self.infinity_io_flag)
-                    df.loc[fixme, f] = -float("inf")
-                elif utils.numericish(self._none_as_infinity_bias(t, f)):
+                    if fixme.any():
+                        df.loc[fixme, f] = -float("inf")
+            for f in all_fields:
+                if not utils.numericish(self.infinity_io_flag) and utils.numericish(self._none_as_infinity_bias(t, f)):
                     assert self.infinity_io_flag is None
                     df[f].fillna(value=self._none_as_infinity_bias(t, f) * float("inf"), inplace=True)
                 dt = self.data_types.get(t, {}).get(f, None)
@@ -301,7 +470,7 @@ class PanDatFactory(object):
                 if number_v is not None and safe_apply(int)(number_v) == number_v:
                     number_v = int(number_v)
                 return value if number_v is None else number_v
-            dat.parameters[val_fld] = _faster_df_apply(dat.parameters, lambda row: fix_value(row))
+            dat.parameters[val_fld] = utils.faster_df_apply(dat.parameters, lambda row: fix_value(row))
         return dat
     def _pre_write_adjustment(self, dat):
         '''
@@ -320,23 +489,40 @@ class PanDatFactory(object):
         rtn = self.copy_pan_dat(rtn) # deep copy so data changes don't side effect
         if self.parameters: # Assuming a parameters table without parameters specification is just a naive developer
             fld = self.data_fields["parameters"][0]
-            rtn.parameters[fld] = _faster_df_apply(rtn.parameters,
-                                                   lambda row: None if isnull(row[fld]) else str(row[fld]))
+            rtn.parameters[fld] = utils.faster_df_apply(rtn.parameters,
+                                                   lambda row: None if isnull(row[fld]) else row[fld])
         if self.infinity_io_flag == "N/A":
             return rtn
-        apply = _faster_df_apply
+        apply = utils.faster_df_apply
         for t in set(self.all_tables).difference(["parameters"]): # parameters table is handled differently
             df = getattr(rtn, t)
-            for f in self.primary_key_fields.get(t, ()) + self.data_fields.get(t, ()):
-                if utils.numericish(self.infinity_io_flag):
+            all_fields = tuple(self.primary_key_fields.get(t, ()) + self.data_fields.get(t, ()))
+            fields_w_issues = set()
+            if utils.numericish(self.infinity_io_flag):
+                def find_fields_w_issues(row):
+                    for f in all_fields:
+                        if utils.numericish(row[f]) and abs(row[f]) >= self.infinity_io_flag:
+                            fields_w_issues.add(f)
+                apply(df, find_fields_w_issues)
+                for f in fields_w_issues:
                     fixme = apply(df, lambda row: utils.numericish(row[f]) and row[f] >= self.infinity_io_flag)
-                    df.loc[fixme, f] = self.infinity_io_flag
+                    if fixme.any():
+                        df.loc[fixme, f] = self.infinity_io_flag
                     fixme = apply(df, lambda row: utils.numericish(row[f]) and row[f] <= -self.infinity_io_flag)
-                    df.loc[fixme, f] = -self.infinity_io_flag
-                elif utils.numericish(self._none_as_infinity_bias(t, f)):
+                    if fixme.any():
+                        df.loc[fixme, f] = -self.infinity_io_flag
+            else:
+                all_fields = tuple(f for f in all_fields if utils.numericish(self._none_as_infinity_bias(t, f)))
+                def find_fields_w_issues(row):
+                    for f in all_fields:
+                        if row[f] == float("inf") * self._none_as_infinity_bias(t, f):
+                            fields_w_issues.add(f)
+                apply(df, find_fields_w_issues)
+                for f in fields_w_issues:
                     assert self.infinity_io_flag is None
                     fixme = apply(df, lambda row: row[f] == float("inf") * self._none_as_infinity_bias(t, f))
-                    df.loc[fixme, f] = None
+                    if fixme.any():
+                        df.loc[fixme, f] = None
         return rtn
     def set_data_type(self, table, field, number_allowed = True,
                       inclusive_min = True, inclusive_max = False, min = 0, max = float("inf"),
@@ -471,6 +657,17 @@ class PanDatFactory(object):
             predicate_name = next(i for i in count() if i not in self._data_row_predicates[table])
         self._data_row_predicates[table][predicate_name] = RowPredicateInfo(predicate, predicate_kwargs_maker,
                                                                             predicate_failure_response)
+    def get_row_predicates(self, table):
+        '''
+        return all the row predicates for a given table
+
+        :param table: a table in the schema
+
+        :return: a dictionary mapping predicate_name to RowPredicateInfo named tuple (the entries of which
+                 are based on the prior call to add_data_row_predicate).
+        '''
+        verify(table in self.all_tables, "Unrecognized table name %s"%table)
+        return {k: v for k, v in self._data_row_predicates.get(table, {}).items()}
 
     def add_parameter(self, name, default_value, number_allowed = True,
                       inclusive_min = True, inclusive_max = False, min = 0, max = float("inf"),
@@ -520,7 +717,7 @@ class PanDatFactory(object):
         verify(self.data_fields["parameters"][0] not in self._data_types.get("parameters", {}),
                 "Don't set the data type for the parameters data field if you are going to use add_parameters.")
         verify(not self._has_been_used,
-               "The parameters can't be changed after a TicDatFactory has been used.")
+               "The parameters can't be changed after a PanDatFactory has been used.")
         td = None
         if enforce_type_rules:
             td = TypeDictionary.safe_creator(number_allowed, inclusive_min, inclusive_max,
@@ -529,6 +726,17 @@ class PanDatFactory(object):
         ParameterInfo = clt.namedtuple("ParameterInfo", ["type_dictionary", "default_value"])
         self._parameters[name] = ParameterInfo(td, default_value)
 
+    def remove_parameter(self, name):
+        '''
+        Undo a previous call to add_parameter.
+
+        :param name: name of the parameter to remove
+
+        :return:
+        '''
+        verify(name in self._parameters, f"{name} is not a valid parameter")
+        self._parameters.pop(name)
+
     def set_default_value(self, table, field, default_value):
         """
         sets the default value for a specific field
@@ -708,21 +916,31 @@ class PanDatFactory(object):
                 self._default_values[tbl][fld] = 0
         self._data_types = clt.defaultdict(dict)
         self._data_row_predicates = clt.defaultdict(dict)
+        self._tooltips = {}
         self._foreign_keys = clt.defaultdict(set)
         self._parameters = {}
         self._infinity_io_flag = ["N/A"]
+        self._xlsx_trailing_empty_rows = ["prune"]
+        self._duplicates_ticdat_init = ["assert"]
         self._none_as_infinity_bias_cache = {}
 
 
         self.all_tables = frozenset(init_fields)
         superself = self
         class PanDat(object):
+            def _len_dict(self):
+                '''
+                :return: a dictionary summarizing table lengths. Zero length tables omitted. Safe to use, I won't change
+                '''
+                return {t: l for t in superself.all_tables for l in [len(getattr(self, t))] if l}
             def __repr__(self):
                 tlen = lambda t: len(getattr(self, t)) if isinstance(getattr(self, t), DataFrame) else None
                 return "pd: {" + ", ".join("%s: %s"%(t, tlen(t)) for t in sorted(superself.all_tables)) + "}"
             def __init__(self, **init_tables):
                 superself._trigger_has_been_used()
-                for t in init_tables :
+                init_tables = {k: v for k,v in init_tables.items() if isinstance(v, (pd.DataFrame, pd.Series)) or
+                               utils.safe_apply(bool)(v)}
+                for t in init_tables:
                     verify(t in superself.all_tables, "Unexpected table name %s"%t)
                     tbl = safe_apply(DataFrame)(init_tables[t])
                     if tbl is None and dictish(init_tables[t]) and all(map(stringish, init_tables[t])):
@@ -731,10 +949,18 @@ class PanDatFactory(object):
                            "Failed to provide a valid DataFrame or DataFrame construction argument for %s"%t)
                     setattr(self, t, tbl.copy())
                     df = getattr(self, t)
+                    pks = superself.primary_key_fields.get(t, ())
+                    if pks and not set(pks).intersection(df.columns) and \
+                       set(pks) == utils.safe_apply(lambda: set(df.index.names))():
+                        df.reset_index(drop=False, inplace=True)
                     if list(df.columns) == list(range(len(df.columns))) and \
                        len(df.columns) >= len(superself._all_fields(t)):
                         df.rename(columns={f1:f2 for f1, f2 in zip(df.columns, superself._all_fields(t))},
                                   inplace=True)
+                    if list(df.columns) != list(range(len(df.columns))):
+                        for f, d in superself.default_values.get(t, {}).items():
+                            if f not in df.columns:
+                                df[f] = d
                 for t in set(superself.all_tables).difference(init_tables):
                     setattr(self, t, DataFrame({f:[] for f in utils.all_fields(superself, t)}))
                 missing_fields = {(t, f) for t in superself.all_tables for f in superself._all_fields(t)
@@ -814,8 +1040,6 @@ class PanDatFactory(object):
         msg = []
         verify(self.good_pan_dat_object(pan_dat, msg.append),
                "pan_dat not a good object for this factory : %s"%"\n".join(msg))
-        if self.find_duplicates(pan_dat):
-            print("--> Warning: duplicate rows will force copy_to_tic_dat to return an object with fewer total rows")
         rtn = self._copy_to_tic_dat(pan_dat)
         from ticdat import TicDatFactory
         # note - this has a minor problem in that it might downcast pandas.Timestamp to numpy.datetime64
@@ -829,10 +1053,9 @@ class PanDatFactory(object):
                 sch[t] = [[], list(getattr(pan_dat, t).columns)]
         from ticdat import TicDatFactory
         tdf = TicDatFactory(**sch)
+        tdf.set_duplicates_ticdat_init(self.duplicates_ticdat_init)
         def df(t):
             rtn = getattr(pan_dat, t)
-            if self.primary_key_fields.get(t, ()):
-                return rtn.set_index(list(self.primary_key_fields[t]), drop=False)
             if t in self.generic_tables and not keep_generics_as_df:
                 return list(map(list, rtn.itertuples(index=False)))
             return rtn
@@ -910,7 +1133,7 @@ class PanDatFactory(object):
                 def bad_row(row):
                     data = row[field]
                     return not data_type.valid_data(None if isnull(data) else data)
-                where_bad_rows = _faster_df_apply(_table, bad_row, trip_wire_check=check_too_many)
+                where_bad_rows = utils.faster_df_apply(_table, bad_row, trip_wire_check=check_too_many)
                 if where_bad_rows.any():
                     rtn[TableField(table, field)] = _table[where_bad_rows].copy() if as_table else where_bad_rows
                 if number_failures[0] >= max_failures:
@@ -986,7 +1209,7 @@ class PanDatFactory(object):
               "Unhandled": Exceptions resulting from calling a row predicate (or a predicate_kwargs_maker) will not be
                            handled by data_row_failures.
               "__debug__": Since "Handled as Failure" makes more sense for production runs and "Unhandled" makes more
-                           sense for debugging, this value will use the latter if __debug__ is True and the former
+                           sense for debugging, this option will use the latter if __debug__ is True and the former
                            otherwise. See -o and __debug__ in Python documentation for more details.
 
         :param max_failures: number. An upper limit on the number of failures to find. Will short circuit and return
@@ -1012,15 +1235,15 @@ class PanDatFactory(object):
         number_failures = [0]
         check_too_many_bool = check_too_many_msg = None
         if max_failures < float("inf"):
-            def check_too_many_bool(is_bad_row):  # here _faster_df_apply is applying a function that
+            def check_too_many_bool(is_bad_row):  # here faster_df_apply is applying a function that
                 if is_bad_row:  # returns True when the row is bad and False o/wise (the boolean reverse of the
                     number_failures[0] += 1 # of the row predicate)
                     if number_failures[0] >= max_failures:  # all future rows will be good
                         return lambda row: False  # which in this context is not bad (i.e. False)
 
-            def check_too_many_msg(true_or_msg):  # here _faster_df_apply is applying a function that
+            def check_too_many_msg(true_or_msg):  # here faster_df_apply is applying a function that
                 if true_or_msg is not True:  # returns True when the row is good and a string o/wise
-                    number_failures[0] += 1  # (i.e. _faster_df_applyu is using the actual row predicate)
+                    number_failures[0] += 1  # (i.e. faster_df_applyu is using the actual row predicate)
                     if number_failures[0] >= max_failures:  # all future rows will be good
                         return lambda row: True  # which in this context is True
 
@@ -1080,7 +1303,7 @@ class PanDatFactory(object):
                         bad_row = (lambda row: not rpi.predicate(row, **predicate_kwargs)) \
                                   if exception_handling == "Unhandled" else (lambda row: not _p(row))
 
-                        where_bad_rows = _faster_df_apply(_table, bad_row, trip_wire_check=check_too_many_bool)
+                        where_bad_rows = utils.faster_df_apply(_table, bad_row, trip_wire_check=check_too_many_bool)
                         if where_bad_rows.any():
                             rtn[TPN(tbl, pn)] = _table[where_bad_rows].copy() if as_table else where_bad_rows
                     else:
@@ -1091,7 +1314,7 @@ class PanDatFactory(object):
                                 return f"Exception<{e}>"
                         predicate = (lambda row: rpi.predicate(row, **predicate_kwargs)) \
                                     if exception_handling == "Unhandled" else (lambda row: _p(row))
-                        predicate_result = _faster_df_apply(_table, predicate, trip_wire_check=check_too_many_msg)
+                        predicate_result = utils.faster_df_apply(_table, predicate, trip_wire_check=check_too_many_msg)
                         where_bad_rows = predicate_result.apply(lambda x: x is not True)
                         if where_bad_rows.any():
                             if as_table:
@@ -1191,8 +1414,8 @@ class PanDatFactory(object):
                 # I'm casting to list here because child is a copy that doesn't have the original index
                 # This is fixable, (i.e. we could return a Series with an index matching the original child table)
                 # but the fix isn't high priority.
-                rtn[fk] = list(_faster_df_apply(child, lambda row: row[magic_field*2] in bad_rows,
-                                                trip_wire_check=check_too_many))
+                rtn[fk] = list(utils.faster_df_apply(child, lambda row: row[magic_field*2] in bad_rows,
+                                                     trip_wire_check=check_too_many))
                 if number_failures[0] >= max_failures:
                     return rtn
         return rtn
