@@ -143,8 +143,10 @@ class _PostgresFactory(freezable_factory(object, "_isFrozen"),):
             rtn = self.tdf.default_values[t][f]
             if forced_field_types.get((t, f)) in ("bool", "boolean"):
                 return bool(rtn)
-            if rtn is None:
+            if rtn is None or rtn == "":
                 return "NULL"
+            if stringish(rtn) and rtn:
+                return f"'{rtn}'"
             return rtn
 
         def nullable(t, f):
@@ -157,7 +159,7 @@ class _PostgresFactory(freezable_factory(object, "_isFrozen"),):
 
         def default_sql_str(t, f):
             fld_type = self.tdf.data_types.get(t, {}).get(f)
-            if fld_type and fld_type.datetime:
+            if (fld_type and fld_type.datetime) or get_fld_type(t, f, '') == "bytea":
                 return ""
             return f" DEFAULT {db_default(t, f)}"
 
@@ -196,7 +198,7 @@ class _PostgresFactory(freezable_factory(object, "_isFrozen"),):
         all_fields = lambda t: self.tdf.primary_key_fields.get(t, ()) + self.tdf.data_fields.get(t, ())
         good_forced_field_type_entry = lambda k, v: isinstance(k, tuple) and len(k) == 2 \
                         and k[1] in all_fields(k[0]) and v in \
-                        ["text", "integer", "float", "bool", "boolean", "timestamp", "date"]
+                        ["text", "integer", "float", "bool", "boolean", "timestamp", "date", "bytea"]
         verify(dictish(forced_field_types) and
                all(good_forced_field_type_entry(k, v) for k,v in forced_field_types.items()),
                "bad forced_field_types argument")
@@ -501,7 +503,8 @@ class PostgresPanFactory(_PostgresFactory):
         assert self.tdf.good_pan_dat_object(rtn, msg.append), str(msg)
         return self.tdf._general_post_read_adjustment(rtn, push_parameters_to_be_valid=True)
 
-    def write_data(self, pan_dat, engine, schema, pre_existing_rows=None, active_fld=""):
+    def write_data(self, pan_dat, engine, schema, pre_existing_rows=None, active_fld="",
+                   progress=None):
         '''
         write the PanDat data to a postgres database
 
@@ -518,6 +521,8 @@ class PostgresPanFactory(_PostgresFactory):
                            Must be compliant w PG naming conventions, which are different from ticdat field naming
                            conventions. Typically developer can ignore this argument, designed for expert support.
 
+        :param progress: if provided, a ticdat.Progress object that is called every time a table is uploaded
+
         :return:
         '''
         verify(_pg_name(active_fld) ==  active_fld, "active_fld needs to be compliant with PG naming conventions")
@@ -529,17 +534,13 @@ class PostgresPanFactory(_PostgresFactory):
         self.check_tables_fields(engine, schema, error_on_missing_table=True) # call self.write_schema as needed
         self._handle_prexisting_rows(engine, schema, pre_existing_rows or {})
         pan_dat = self.tdf._pre_write_adjustment(pan_dat)
-        for table in self._ordered_tables():
+        to_upload = self._ordered_tables()
+        for i, table in enumerate(to_upload):
             df = getattr(pan_dat, table).copy(deep=True)
             fields = self.tdf.primary_key_fields.get(table, ()) + self.tdf.data_fields.get(table, ())
             df.rename(columns={f: _pg_name(f) for f in fields}, inplace=True)
             if table in active_field_tables:
                 df[active_fld] = True
             df.to_sql(name=table, schema=schema, con=engine, if_exists="append", index=False)
-
-
-
-
-
-
-
+            if progress and not progress.numerical_progress("Uploading...", 100.*i/len(to_upload)):
+                break
