@@ -2,10 +2,11 @@ import ticdat.utils as utils
 from ticdat import TicDatFactory, PanDatFactory
 from ticdat.testing.ticdattestutils import pan_dat_maker
 from ticdat.pgtd import _can_unit_test, PostgresTicFactory, PostgresPanFactory, _pg_name
-from ticdat.testing.ticdattestutils import flagged_as_run_alone, fail_to_debugger
+from ticdat.testing.ticdattestutils import flagged_as_run_alone, fail_to_debugger, memo
 import time
 import datetime
 import math
+import pickle
 
 import unittest
 try:
@@ -21,7 +22,7 @@ try:
 except:
     pd = None
 try:
-    import dateutil
+    import dateutil, dateutil.parser
 except:
     dateutil = None
 
@@ -191,12 +192,9 @@ class TestPostres(unittest.TestCase):
         pgtf.write_data(pan_dat, self.engine, test_schema)
         pg_pan_dat = pgtf.create_pan_dat(self.engine, test_schema)
         self.assertTrue(pdf._same_data(pan_dat, pg_pan_dat))
-        from ticdat.pandatfactory import _faster_df_apply
-        pan_dat.categories["Max Nutrition"] = _faster_df_apply(pan_dat.categories,
+        pan_dat.categories["Max Nutrition"] = utils.faster_df_apply(pan_dat.categories,
                                                                lambda row: numpy.int64(row["Max Nutrition"]))
-        pan_dat.foods["Cost"] = _faster_df_apply(pan_dat.foods, lambda row: numpy.float64(row["Cost"]))
-        from framework_utils.helper_utils import memo
-        memo(pan_dat)
+        pan_dat.foods["Cost"] = utils.faster_df_apply(pan_dat.foods, lambda row: numpy.float64(row["Cost"]))
         pgtf.write_data(pan_dat, self.engine, test_schema)
         pg_pan_dat = pgtf.create_pan_dat(self.engine, test_schema)
         self.assertTrue(pdf._same_data(pan_dat, pg_pan_dat))
@@ -309,11 +307,11 @@ class TestPostres(unittest.TestCase):
             big_dat.categories[str(k)] = [0,100]
         pgtf.write_schema(self.engine, test_schema)
         pgtf.write_data(big_dat, self.engine, test_schema, dsn=self.postgresql.dsn())
-        print(f"****{time.time()-now}****")
+        print(f"\n**** tdf writing {big_dat._len_dict()} : {time.time()-now}****\n")
         now = time.time()
         self.assertFalse(pgtf.find_duplicates(self.engine, test_schema))
         pg_tic_dat = pgtf.create_tic_dat(self.engine, test_schema)
-        print(f"****{time.time()-now}****")
+        print(f"**** tdf reading {big_dat._len_dict()} : {time.time()-now}****")
         self.assertTrue(diet_schema._same_data(big_dat, pg_tic_dat))
 
         med_dat = diet_schema.copy_tic_dat(diet_dat)
@@ -335,11 +333,11 @@ class TestPostres(unittest.TestCase):
             big_dat.categories[str(k)] = [0,100]
         pgtf.write_schema(self.engine, test_schema)
         pgtf.write_data(big_dat, self.engine, test_schema, dsn=self.postgresql.url())
-        print(f"**!!*{time.time()-now}*!!   *")
+        print(f"\ntdf writing {big_dat._len_dict()} {time.time()-now}*!!*\n")
         now = time.time()
         self.assertFalse(pgtf.find_duplicates(self.engine, test_schema))
         pg_tic_dat = pgtf.create_tic_dat(self.engine, test_schema)
-        print(f"****{time.time()-now}****")
+        print(f"\ntdf reading and dup {big_dat._len_dict()} {time.time()-now}****\n")
         self.assertTrue(diet_schema._same_data(big_dat, pg_tic_dat))
 
         med_dat = diet_schema.copy_tic_dat(diet_dat)
@@ -419,6 +417,30 @@ class TestPostres(unittest.TestCase):
         dups = tdf.pgsql.find_duplicates(self.engine, test_schema)
         self.assertTrue(dups == {'three': {(1, 2, 2): 2}, 'two': {(1, 2): 3}, 'one': {1: 3, 2: 2}})
 
+    def test_pd_progress(self):
+        if not self.can_run:
+            return
+        schema = "t_pg_d_pd"
+        tdf = diet_schema
+        pdf = PanDatFactory.create_from_full_schema(tdf.schema(include_ancillary_info=True))
+        pdf.set_infinity_io_flag(1e12)
+        pgpf = pdf.pgsql
+        pan_dat = pan_dat_maker(tdf.schema(), diet_dat)
+        pgpf.write_schema(self.engine, schema, include_ancillary_info=False)
+        class MyProgress(utils.Progress):
+            def __init__(self):
+                self.uploaded=0
+                super().__init__(quiet=True)
+            def numerical_progress(self, theme, progress):
+                self.uploaded+=1
+                return self.uploaded < 2
+        progress = MyProgress()
+        pgpf.write_data(pan_dat, self.engine, schema, progress=progress)
+        self.assertTrue(progress.uploaded == 2)
+        pg_pan_dat = pgpf.create_pan_dat(self.engine, schema)
+        self.assertFalse(pdf._same_data(pan_dat, pg_pan_dat))
+        self.assertTrue(pg_pan_dat._len_dict() == {'categories': 4, 'foods': 9})
+
     def test_diet_pd(self):
         if not self.can_run:
             return
@@ -492,14 +514,15 @@ class TestPostres(unittest.TestCase):
         now = time.time()
         pgpf.write_schema(self.engine, schema)
         pgpf.write_data(pan_dat, self.engine, schema)
-        print(f"**&&*{time.time()-now}**&&**")
+        print(f"\npdf writing {big_dat._len_dict()} : {time.time()-now}**&&**\n")
         now = time.time()
         pg_pan_dat = pgpf.create_pan_dat(self.engine, schema)
-        print(f"*&&*{time.time()-now}**&&**")
+        print(f"\npdf reading {big_dat._len_dict()} : {time.time()-now}**&&**\n")
         self.assertTrue(pdf._same_data(pan_dat, pg_pan_dat))
 
     def test_extra_fields_pd(self):
         pdf = PanDatFactory(boger = [["a"], ["b", "c"]])
+        pdf.set_default_value("boger", "c", "hoya")
         dat = pdf.PanDat(boger = pd.DataFrame({"a": [1, 2, 3], "b":[4, 5,6], "c":['a', 'b', 'c']}))
         schema = "test_pd_extra_fields"
         pdf.pgsql.write_schema(self.engine, schema, forced_field_types={("boger", "c"): "text",
@@ -512,7 +535,32 @@ class TestPostres(unittest.TestCase):
         pdf2.pgsql.write_data(dat2_2, self.engine, schema)
         dat = pdf.pgsql.create_pan_dat(self.engine, schema)
         self.assertTrue(list(dat.boger["a"]) == [10, 300] and list(dat.boger["b"]) == [40, 60])
-        self.assertTrue(len(set(dat.boger["c"])) == 1)
+        self.assertTrue(set(dat.boger["c"]) == {'hoya'})
+
+        pdf2 = PanDatFactory(boger=[["b"], ["a"]])
+        dat2_2 = pdf2.PanDat(boger = pd.DataFrame({"b":[41, 61], "a": [21, 31]}))
+        pdf2.pgsql.write_data(dat2_2, self.engine, schema)
+        dat = pdf.pgsql.create_pan_dat(self.engine, schema)
+        self.assertTrue(list(dat.boger["a"]) == [21, 31] and list(dat.boger["b"]) == [41, 61])
+        self.assertTrue(set(dat.boger["c"]) == {'hoya'})
+
+        pdf2 = PanDatFactory(boger=[["b"], ["z", "a"]])
+        dat2_2 = pdf2.PanDat(boger = pd.DataFrame({"z": ["this", "that", "other"], "b":[42, 61, 1], "a": [22, 31, 2]}))
+        self.assertTrue(list(dat2_2.boger.columns) == ['b', 'z', 'a'])
+        # extra columns won't work when they are part of the PanDatFactory!
+        ex = []
+        try:
+            pdf2.pgsql.write_data(dat2_2, self.engine, schema)
+        except Exception as e:
+            ex.append(str(e))
+        self.assertTrue("Unable to recognize" in ex[0])
+
+        pdf2 = PanDatFactory(boger=[["b"], ["a"]])
+        pdf2.pgsql.write_data(dat2_2, self.engine, schema)
+        dat = pdf.pgsql.create_pan_dat(self.engine, schema)
+        self.assertTrue(list(dat.boger["a"]) == [22, 31, 2] and list(dat.boger["b"]) == [42, 61, 1])
+        self.assertTrue(set(dat.boger["c"]) == {'hoya'})
+
 
     def test_time_stamp(self):
         tdf = TicDatFactory(table=[["Blah"],["Timed Info"]])
@@ -843,6 +891,57 @@ class TestPostres(unittest.TestCase):
         dat_3 = pdf.copy_to_tic_dat(pan_dat)
         self.assertTrue(tdf._same_data(dat, dat_3, nans_are_same_for_data_rows=True))
 
+    def test_pgtd_pandas_none_null(self):
+        # roundoff (issue 5) is particularly interested in getting the PG NULL to map onto None with PanDatFactory
+        if not self.can_run:
+            return
+        schema = test_schema+"_none_null"
+        tdf_1 = TicDatFactory(t_one=[["Field One"], ["Field Two"]])
+        tdf_1.set_data_type("t_one", "Field Two", strings_allowed='*', number_allowed=False, nullable=True)
+        dat = tdf_1.TicDat(t_one = [["a", None], ["b", "b"], [None, "c"], ["c", "c"]])
+        self.assertTrue(len(dat.t_one) == 4)
+        tdf_1.pgsql.write_schema(self.engine, schema, include_ancillary_info=False)
+        tdf_1.pgsql.write_data(dat, self.engine, schema)
+        dat_2 = tdf_1.pgsql.create_tic_dat(self.engine, schema)
+        self.assertTrue(tdf_1._same_data(dat, dat_2))
+        pdf = PanDatFactory.create_from_full_schema(tdf_1.schema(include_ancillary_info=True))
+        pan_dat = pdf.pgsql.create_pan_dat(self.engine, schema)
+        dat_3 = pdf.copy_to_tic_dat(pan_dat)
+        self.assertTrue(tdf_1._same_data(dat, dat_3))
+        self.assertTrue((dat_3.t_one["a"]['Field Two'] == None) and None in dat_3.t_one)
+
+    def test_pickle_w_text_columns(self):
+        # this text in response to a request for blob field type support. I'd rather not support blobs
+        # if I don't have to, but supporting pickle.dumps is legit
+        if not self.can_run:
+            return
+        schema = test_schema+"_pickle"
+        tdf = TicDatFactory(t_one=[["Field One"], ["Field Two"]])
+        dat = tdf.TicDat(t_one = [["a", pickle.dumps([1, 2])], ["b", pickle.dumps({"a": [1, 2], "b": [1, 2, 3]})]])
+        pdf = tdf.clone(clone_factory=PanDatFactory)
+        pan_dat = pdf.PanDat(**{t: getattr(tdf.copy_to_pandas(dat, reset_index=True), t) for t in tdf.all_tables})
+        self.assertTrue(len(dat.t_one) == len(pan_dat.t_one) == 2)
+        tdf.pgsql.write_schema(self.engine, schema, include_ancillary_info=False,
+                               forced_field_types={("t_one", "Field Two"): "bytea"})
+        tdf.pgsql.write_data(dat, self.engine, schema)
+        dat_2 = tdf.pgsql.create_tic_dat(self.engine, schema)
+        self.assertFalse(tdf._same_data(dat, dat_2)) # its  weird this way, but no matter, we only care about pickle
+        self.assertTrue(set(dat.t_one) == set(dat_2.t_one))
+        for k, r in dat.t_one.items():
+            self.assertTrue(pickle.loads(r["Field Two"]) == pickle.loads(dat_2.t_one[k]["Field Two"]))
+
+        schema = test_schema+"_pickle_pdf"
+        pdf.pgsql.write_schema(self.engine, schema, include_ancillary_info=False,
+                               forced_field_types={("t_one", "Field Two"): "bytea"})
+        pdf.pgsql.write_data(pan_dat, self.engine, schema)
+        pan_dat_2 = pdf.pgsql.create_pan_dat(self.engine, schema)
+        dat_3 = pdf.copy_to_tic_dat(pan_dat)
+        dat_4 = pdf.copy_to_tic_dat(pan_dat_2)
+        self.assertFalse(tdf._same_data(dat_3, dat_4))
+        self.assertTrue(set(dat_3.t_one) == set(dat_4.t_one))
+        for k, r in dat_3.t_one.items():
+            self.assertTrue(pickle.loads(r["Field Two"]) == pickle.loads(dat_4.t_one[k]["Field Two"]))
+
 test_schema = 'test'
 
 
