@@ -1,11 +1,11 @@
 import unittest
-from ticdat.pandatfactory import PanDatFactory
+from ticdat.pandatfactory import PanDatFactory, remove_trailing_all_nan
 from ticdat.utils import DataFrame, numericish, ForeignKey, ForeignKeyMapping
 import ticdat.utils as utils
 from ticdat.testing.ticdattestutils import fail_to_debugger, flagged_as_run_alone, netflowPandasData
 from ticdat.testing.ticdattestutils import netflowSchema, copy_to_pandas_with_reset, dietSchema, netflowData
 from ticdat.testing.ticdattestutils import addNetflowForeignKeys, sillyMeSchema, dietData, pan_dat_maker
-from ticdat.testing.ticdattestutils import addDietForeignKeys
+from ticdat.testing.ticdattestutils import addDietForeignKeys, dietData
 from ticdat.ticdatfactory import TicDatFactory
 import itertools
 from math import isnan
@@ -22,6 +22,91 @@ def _deep_anonymize(x)  :
 class TestUtils(unittest.TestCase):
     canRun = False
 
+    def test_advanced_kwargs_cloning(self):
+        if not self.canRun:
+            return
+        tdf = TicDatFactory(foo=[["Primary Key Field"], ["Data Field"]])
+        tdf.set_data_type("foo", "Primary Key Field", number_allowed=True, min=0, max=float("inf"), inclusive_min=True,
+                          inclusive_max=False)
+        tdf.add_data_row_predicate("foo", lambda row: True, predicate_name="normal")
+        tdf.add_data_row_predicate("foo", lambda row, dummy: dummy, predicate_name="advanced one",
+                                   predicate_kwargs_maker=lambda dat: {"dummy": True})
+        tdf.add_data_row_predicate("foo", lambda row, min_pk: row["Primary Key Field"] != min_pk+1,
+                                   predicate_name="advanced two",
+                                   predicate_kwargs_maker=lambda dat: {"min_pk": min(dat.foo)})
+        tdf.clone()
+        dat_one = tdf.TicDat(foo=[[1, "junk"], [12, "junk"]])
+        tdf.freeze_me(dat_one)
+        dat_two = tdf.TicDat(foo=[[1, "junk"], [2, "junk"]])
+        tdf.freeze_me(dat_two)
+        self.assertFalse(tdf.find_data_row_failures(dat_one, exception_handling="Handled as Failure"))
+        self.assertTrue(tdf.find_data_row_failures(dat_two, exception_handling="Handled as Failure"))
+        pdat_one = tdf.copy_to_pandas(dat_one, reset_index=True)
+        pdat_two = tdf.copy_to_pandas(dat_two, reset_index=True)
+        # the advanced row predicates don't copy over when cloning to an explicitly different type
+        pdf = tdf.clone(clone_factory=PanDatFactory)
+        self.assertFalse(pdf.find_data_row_failures(pdat_one, exception_handling="Handled as Failure"))
+        self.assertFalse(pdf.find_data_row_failures(pdat_two, exception_handling="Handled as Failure"))
+        pdf.add_data_row_predicate("foo", lambda row, dummy: dummy, predicate_name="advanced one",
+                                   predicate_kwargs_maker=lambda dat: {"dummy": True})
+        pdf.add_data_row_predicate("foo", lambda row, min_pk: row["Primary Key Field"] != min_pk+1,
+                                   predicate_name="advanced two",
+                                   predicate_kwargs_maker=lambda dat: {"min_pk": min(dat.foo["Primary Key Field"])})
+        pdf = pdf.clone()
+        self.assertFalse(pdf.find_data_row_failures(pdat_one, exception_handling="Handled as Failure"))
+        self.assertTrue(pdf.find_data_row_failures(pdat_two, exception_handling="Handled as Failure"))
+
+        pdf_w_bum_preds = tdf.clone(clone_factory=PanDatFactory.create_from_full_schema)
+        self.assertFalse(any(pdf_w_bum_preds.find_data_type_failures(_) for _ in [pdat_one, pdat_two]))
+        self.assertTrue(all(pdf_w_bum_preds.find_data_row_failures(_, exception_handling="Handled as Failure")
+                            for _ in [pdat_one, pdat_two]))
+        tdf2 = pdf.clone(clone_factory=TicDatFactory)
+        # the advanced row predicates don't copy over when cloning to an explicitly different type
+        self.assertFalse(tdf2.find_data_row_failures(dat_one, exception_handling="Handled as Failure"))
+        self.assertFalse(tdf2.find_data_row_failures(dat_two, exception_handling="Handled as Failure"))
+        tdf_w_bum_preds = pdf.clone(clone_factory=TicDatFactory.create_from_full_schema)
+        self.assertFalse(any(tdf_w_bum_preds.find_data_type_failures(_) for _ in [dat_one, dat_two]))
+        self.assertTrue(all(tdf_w_bum_preds.find_data_row_failures(_, exception_handling="Handled as Failure")
+                            for _ in [dat_one, dat_two]))
+
+
+    def testDefaultAdd(self):
+        if not self.canRun:
+            return
+        tdf = TicDatFactory(**dietSchema())
+        ticDat = tdf.freeze_me(tdf.TicDat(**{t:getattr(dietData(),t) for t in tdf.primary_key_fields}))
+        panDat = pan_dat_maker(dietSchema(), ticDat)
+        tdf2 = TicDatFactory(**{k:[p,d] if k!="foods" else [p, list(d)+["extra"]] for k,(p,d) in dietSchema().items()})
+        pdf2 = tdf2.clone(clone_factory=PanDatFactory)
+        panDat2 = pdf2.PanDat(**{t: getattr(panDat, t) for t in tdf.all_tables})
+        ticDat2 = tdf2.TicDat(**{t: getattr(ticDat, t) for t in tdf.all_tables})
+        self.assertTrue(tdf2._same_data(ticDat2, pdf2.copy_to_tic_dat(panDat2), epsilon=1e-5))
+        self.assertTrue(set(panDat2.foods["extra"]) == {0})
+        pdf3 = pdf2.clone()
+        pdf3.set_default_value("foods", "extra", 100)
+        panDat3 = pdf3.PanDat(**{t: getattr(panDat, t) for t in tdf.all_tables})
+        self.assertFalse(tdf2._same_data(ticDat2, pdf3.copy_to_tic_dat(panDat3), epsilon=1e-5))
+        self.assertTrue(set(panDat3.foods["extra"]) == {100})
+
+    def testCopying(self):
+        tdf = TicDatFactory(**dietSchema())
+        pdf = PanDatFactory(**dietSchema())
+        dat = tdf.copy_tic_dat(dietData())
+        def make_pan_dat(reset_index):
+            rtn = tdf.copy_to_pandas(dat, reset_index=reset_index)
+            return pdf.PanDat(**{k: getattr(rtn, k) for k in tdf.all_tables})
+        dat2 = pdf.copy_to_tic_dat(make_pan_dat(False))
+        dat3 = pdf.copy_to_tic_dat(tdf.copy_to_pandas(dat, reset_index=True))
+        self.assertTrue(tdf._same_data(dat, dat2))
+        self.assertTrue(tdf._same_data(dat, dat3))
+        def make_tic_dat_dat(reset_index):
+            rtn = tdf.copy_to_pandas(dat, reset_index=reset_index)
+            return tdf.TicDat(**{k: getattr(rtn, k) for k in tdf.all_tables})
+        dat2 = make_tic_dat_dat(False)
+        dat3 = make_tic_dat_dat(True)
+        self.assertTrue(tdf._same_data(dat, dat2))
+        self.assertTrue(tdf._same_data(dat, dat3))
+
     def testSimple(self):
         if not self.canRun:
             return
@@ -140,6 +225,7 @@ class TestUtils(unittest.TestCase):
         pdf2.set_default_value("foods", "name", "a")
         pdf2.set_default_value("nutritionQuantities", "food", "a")
         pdf2.replace_data_type_failures(dat_copied)
+        pdf.set_duplicates_ticdat_init("ignore")
         self.assertFalse(pdf._same_data(dat, dat_copied, epsilon=0.00001))
         self.assertFalse(pdf.find_data_type_failures(dat_copied))
         dups = pdf.find_duplicates(dat_copied)
@@ -747,6 +833,40 @@ class TestUtils(unittest.TestCase):
         errs = pdf.find_foreign_key_failures(pan_dat, max_failures=9)
         self.assertTrue(len(errs) == 1 and all(len(_) == 9 for _ in errs.values()))
 
+    def test_trailing_all_nan(self):
+        df = utils.pd.DataFrame({"a": [1, 2, None, None], "b": [10, 20, None, None]})
+        df2 = remove_trailing_all_nan(df)
+        df3 = remove_trailing_all_nan(df2)
+        self.assertTrue([tuple(x) for x in df2.itertuples(index=False)] ==
+                        [tuple(x) for x in df3.itertuples(index=False)] == [(1.0, 10.0), (2.0, 20.0)])
+        df = utils.pd.DataFrame({"a": [1, 2] * 10 + [None, None] * 3, "b": [10, 20] * 10 + [None, float("nan")] * 3})
+        df2 = remove_trailing_all_nan(df)
+        df3 = remove_trailing_all_nan(df2)
+        self.assertTrue([tuple(x) for x in df2.itertuples(index=False)] ==
+                        [tuple(x) for x in df3.itertuples(index=False)] == [(1.0, 10.0), (2.0, 20.0)]*10)
+        df = utils.pd.DataFrame({"a": [1, 2] * 10 + [None] + [2,1] + [None, None] * 3,
+                                 "b": [10, 20] * 10 + [None] + [20,11] + [None, None] * 3})
+        df2 = remove_trailing_all_nan(df)
+        df3 = remove_trailing_all_nan(df2)
+        _tuple = lambda x: tuple(None if utils.pd.isnull(_) else _ for _ in x)
+        self.assertTrue([_tuple(x) for x in df2.itertuples(index=False)] ==
+                        [_tuple(x) for x in df3.itertuples(index=False)] == [(1.0, 10.0), (2.0, 20.0)]*10 +
+                         [(None, None)] + [(2, 20), (1, 11)])
+        df = utils.pd.DataFrame({"a": [1, 2]*1000000 + [None, None]*30, "b": [10, 20]*1000000 +
+                                                                             [float("nan"), None]*30})
+        df2 = remove_trailing_all_nan(df)
+        df3 = remove_trailing_all_nan(df2)
+        self.assertTrue([tuple(x) for x in df2.itertuples(index=False)] ==
+                        [tuple(x) for x in df3.itertuples(index=False)] == [(1.0, 10.0), (2.0, 20.0)]*1000000)
+
+        self.assertTrue(len(remove_trailing_all_nan(utils.pd.DataFrame({"a": [None, float("nan"), None],
+                                                                        "b":[None]*3}))) == 0)
+    def test_empty_maker(self):
+        pdf = PanDatFactory(**dietSchema())
+        dat = pdf.PanDat(nutritionQuantities=[[f"food_{_}", f"cat_{_}", 10] for _ in range(10)],
+                         foods = [[f"food_{_}", 10] for _ in range(1, 0)]) # empty list results
+        self.assertTrue(dat._len_dict() == {"nutritionQuantities": 10})
+
 # Run the tests.
 if __name__ == "__main__":
     if not DataFrame :
