@@ -21,19 +21,69 @@ try:
     import xlsxwriter as xlsx
 except:
     xlsx=None
+try:
+    import openpyxl
+except:
+    openpyxl=None
 
-_can_unit_test = xlrd and xlwt and xlsx
+_the_openpyxl_ext = {".xlsx", ".xlsm"}
+_is_openpyxl_ext = lambda s : any(s.endswith(_) for _ in _the_openpyxl_ext)
+_can_unit_test = xlrd and xlwt and xlsx and openpyxl
 
 # https://github.com/jmcnamara/XlsxWriter/issues/150 ...
 # the xlsxwriter doesn't handle infinity as seamlessly as xls
 _longest_sheet = 30
 
+class _XlrdSheetWrapper(object): # main purpose of this routine is to enforce the use of
+    def __init__(self, sheet, datemode):   # a limited set of xlrd.sheet functions
+        self._sheet = sheet
+        self._datemode = datemode
+    @property
+    def nrows(self):
+        return self._sheet.nrows
+    def row_values(self, row_index):
+        return self._sheet.row_values(row_index)
+    def col_values(self, col_index):
+        return self._sheet.col_values(col_index)
+    def xldate_as_tuple_munge(self, x): # only needed for xlrd
+        rtn = utils.safe_apply(lambda: xlrd.xldate_as_tuple(x, self._datemode))()
+        if rtn is not None:
+            f = datetime.datetime
+            if utils.pd:
+                f = utils.pd.Timestamp
+            return f(year=rtn[0], month=rtn[1], day=rtn[2], hour=rtn[3], minute=rtn[4], second=rtn[5])
+
+class _OpenPyxlSheetWrapper(object):
+    '''
+    This file initially used xlrd for xls and xlsx files. This class allows an openpyxl.sheet to present as a
+    limited xlrd.sheet. Although this choice of abstractions is historical, the resulting code is at least
+    free from lots of "if/else" silliness
+    '''
+    def __init__(self, sheet, prune_trailing_empty_rows):
+        self._sheet = sheet
+        self._max_col = sheet.max_column # for performance reasons, need to capture the range on __init__
+        self._max_row = sheet.max_row
+        if prune_trailing_empty_rows:
+            i = self.nrows-1
+            while i >= 0 and all(x is None for x in self.row_values(i)):
+                i -= 1
+            self._max_row = i+1
+    @property
+    def nrows(self):
+        return self._sheet.max_row-self._sheet.min_row+1
+    def row_values(self, row_index): # openpyxl used 1 based indexing
+        return tuple(x.value for x in next(iter(self._sheet.iter_rows(min_row=row_index+1, max_row=row_index+1,
+                                                                      min_col=1, max_col=self._max_col))))
+    def col_values(self, col_index): # openpyxl used 1 based indexing
+        return tuple(x.value for x in next(iter(self._sheet.iter_cols(min_col=col_index+1, max_col=col_index+1,
+                                                                      min_row=1, max_row=self._max_row))))
+
 class XlsTicFactory(freezable_factory(object, "_isFrozen")) :
     """
     Primary class for reading/writing Excel files with TicDat objects.
-    Your system will need the xlrd package to read .xls and .xlsx files,
-    the xlwt package to write .xls files, and the xlsxwriter package to
-    write .xlsx files.
+    Your system will need the xlrd package to read .xls files, the openpyxl package
+    to read xlsx/xlsm/xltx/xltm files, the xlwt package to write .xls files, and the xlsxwriter
+    package to write .xlsx files.
     Don't create this object explicitly. A XlsTicDatFactory will
     automatically be associated with the xls attribute of the parent
     TicDatFactory.
@@ -47,6 +97,7 @@ class XlsTicFactory(freezable_factory(object, "_isFrozen")) :
         :return:
         """
         self.tic_dat_factory = tic_dat_factory
+        self._dv_dt = {}
         self._isFrozen = True
     def create_tic_dat(self, xls_file_path, row_offsets=None, headers_present = True,
                        treat_inf_as_infinity = True,
@@ -85,7 +136,13 @@ class XlsTicFactory(freezable_factory(object, "_isFrozen")) :
                      is to set must_be_int to true in data_types.
         """
         self._verify_differentiable_sheet_names()
-        verify(xlrd, "xlrd needs to be installed to use this subroutine")
+        verify(utils.safe_apply(os.path.isfile)(xls_file_path), f"{xls_file_path} not a file path")
+        verify(xls_file_path.endswith(".xls") or _is_openpyxl_ext(xls_file_path),
+               f"invalid file extension for {xls_file_path}")
+        if xls_file_path.endswith(".xls"):
+            verify(xlrd, "xlrd needs to be installed to use this subroutine")
+        else:
+            verify(openpyxl, "openpyxl needs to be installed to use this subroutine")
         tdf = self.tic_dat_factory
         verify(not(treat_inf_as_infinity and tdf.generator_tables),
                "treat_inf_as_infinity not implemented for generator tables")
@@ -121,17 +178,23 @@ class XlsTicFactory(freezable_factory(object, "_isFrozen")) :
         verify(utils.stringish(xls_file_path) and os.path.exists(xls_file_path),
                "xls_file_path argument %s is not a valid file path."%xls_file_path)
         try :
-            book = xlrd.open_workbook(xls_file_path)
+            book = xlrd.open_workbook(xls_file_path) if xls_file_path.endswith(".xls") else \
+                openpyxl.load_workbook(xls_file_path, data_only=True)
         except Exception as e:
             raise TicDatError("Unable to open %s as xls file : %s"%(xls_file_path, e))
+        sheet_name = lambda sheet: sheet.name if xls_file_path.endswith(".xls") else sheet.title
         sheets = defaultdict(list)
-        for table, sheet in product(all_tables, book.sheets()) :
-            if table.lower()[:_longest_sheet] == sheet.name.lower().replace(' ', '_')[:_longest_sheet]:
+        book_sheets = lambda: book.sheets() if xls_file_path.endswith(".xls") else book.worksheets
+        for table, sheet in product(all_tables, book_sheets()) :
+            if table.lower()[:_longest_sheet] == sheet_name(sheet).lower().replace(' ', '_')[:_longest_sheet]:
                 sheets[table].append(sheet)
         duplicated_sheets = tuple(_t for _t,_s in sheets.items() if len(_s) > 1)
-        verify(not duplicated_sheets, "The following sheet names were duplicated : " +
+        verify(not duplicated_sheets, "The following sheet names were duplicated s: " +
                ",".join(duplicated_sheets))
-        sheets = FrozenDict({k:v[0] for k,v in sheets.items() })
+        wrapped_sheet = lambda sheet: _XlrdSheetWrapper(sheet, book.datemode) if xls_file_path.endswith(".xls") else \
+                                      _OpenPyxlSheetWrapper(sheet, prune_trailing_empty_rows=
+                                        self.tic_dat_factory.xlsx_trailing_empty_rows == "prune")
+        sheets = FrozenDict({k: wrapped_sheet(v[0]) for k,v in sheets.items()})
         missing_tables = {t for t in all_tables if t not in sheets}
         if missing_tables and print_missing_tables:
             print ("The following table names could not be found in the %s file.\n%s\n"%
@@ -146,12 +209,12 @@ class XlsTicFactory(freezable_factory(object, "_isFrozen")) :
         verify(not any(_ for _ in dup_fields.values()),
                "The following field names were duplicated : \n" +
                "\n".join("%s : "%t + ",".join(bf) for t,bf in dup_fields.items() if bf))
-        return sheets, field_indicies, book.datemode
+        return sheets, field_indicies
     def _create_generator_obj(self, xlsFilePath, table, row_offset, headers_present, treat_inf_as_infinity):
         tdf = self.tic_dat_factory
         ho = 1 if headers_present else 0
         def tableObj() :
-            sheets, field_indicies, datemode = self._get_sheets_and_fields(xlsFilePath,
+            sheets, field_indicies = self._get_sheets_and_fields(xlsFilePath,
                                         (table,), {table:row_offset}, headers_present)
             if table in sheets :
                 sheet = sheets[table]
@@ -159,7 +222,7 @@ class XlsTicFactory(freezable_factory(object, "_isFrozen")) :
                                for field in tdf.data_fields[table])
                 for x in (sheet.row_values(i) for i in range(table_len)[row_offset+ho:]):
                     yield self._sub_tuple(table, tdf.data_fields[table],
-                                          field_indicies[table], treat_inf_as_infinity, datemode)(x)
+                                          field_indicies[table], treat_inf_as_infinity, sheet)(x)
         return tableObj
 
     def _create_tic_dat_dict(self, xls_file_path, row_offsets, headers_present, treat_inf_as_infinity):
@@ -171,7 +234,7 @@ class XlsTicFactory(freezable_factory(object, "_isFrozen")) :
         row_offsets = dict({t:0 for t in self.tic_dat_factory.all_tables}, **row_offsets)
         tdf = self.tic_dat_factory
         rtn = {}
-        sheets, field_indicies, dm = self._get_sheets_and_fields(xls_file_path,
+        sheets, field_indicies = self._get_sheets_and_fields(xls_file_path,
                                     set(tdf.all_tables).difference(tdf.generator_tables),
                                     row_offsets, headers_present, print_missing_tables=True)
         ho = 1 if headers_present else 0
@@ -182,14 +245,14 @@ class XlsTicFactory(freezable_factory(object, "_isFrozen")) :
             table_len = min(len(sheet.col_values(indicies[field]))
                             for field in (fields or indicies))
             if tdf.primary_key_fields.get(tbl, ()) :
-                tableObj = {self._sub_tuple(tbl, tdf.primary_key_fields[tbl], indicies, tiai, dm)(x):
-                            self._sub_tuple(tbl, tdf.data_fields.get(tbl, ()), indicies, tiai, dm)(x)
+                tableObj = {self._sub_tuple(tbl, tdf.primary_key_fields[tbl], indicies, tiai, sheet)(x):
+                            self._sub_tuple(tbl, tdf.data_fields.get(tbl, ()), indicies, tiai, sheet)(x)
                             for x in (sheet.row_values(i) for i in
                                         range(table_len)[row_offsets[tbl]+ho:])}
             elif tbl in tdf.generic_tables:
                 tableObj = None # will be read via PanDatFactory
             else :
-                tableObj = [self._sub_tuple(tbl, tdf.data_fields.get(tbl, ()), indicies, tiai, dm)(x)
+                tableObj = [self._sub_tuple(tbl, tdf.data_fields.get(tbl, ()), indicies, tiai, sheet)(x)
                             for x in (sheet.row_values(i) for i in
                                         range(table_len)[row_offsets[tbl]+ho:])]
             if tableObj is not None:
@@ -224,7 +287,10 @@ class XlsTicFactory(freezable_factory(object, "_isFrozen")) :
                  Row counts smaller than 2 are pruned off, as they aren't duplicates
         """
         self._verify_differentiable_sheet_names()
-        verify(xlrd, "xlrd needs to be installed to use this subroutine")
+        if xls_file_path.endswith(".xls"):
+            verify(xlrd, "xlrd needs to be installed to use this subroutine")
+        else:
+            verify(openpyxl, "openpyxl needs to be installed to use this subroutine")
         verify(utils.dictish(row_offsets) and
                set(row_offsets).issubset(self.tic_dat_factory.all_tables) and
                all(utils.numericish(x) and (x>=0) for x in row_offsets.values()),
@@ -233,7 +299,7 @@ class XlsTicFactory(freezable_factory(object, "_isFrozen")) :
         tdf = self.tic_dat_factory
         pk_tables = tuple(t for t,_ in tdf.primary_key_fields.items() if _)
         rtn = {t:defaultdict(int) for t in pk_tables}
-        sheets, fieldIndicies, dm = self._get_sheets_and_fields(xls_file_path, pk_tables,
+        sheets, fieldIndicies  = self._get_sheets_and_fields(xls_file_path, pk_tables,
                                         row_offsets, headers_present)
         ho = 1 if headers_present else 0
         for table, sheet in sheets.items() :
@@ -242,21 +308,27 @@ class XlsTicFactory(freezable_factory(object, "_isFrozen")) :
             table_len = min(len(sheet.col_values(indicies[field])) for field in fields)
             for x in (sheet.row_values(i) for i in range(table_len)[row_offsets[table]+ho:]) :
                 rtn[table][self._sub_tuple(table, tdf.primary_key_fields[table],
-                                           indicies, treat_inf_as_infinity=True, datemode=dm)(x)] += 1
+                                           indicies, treat_inf_as_infinity=True, sheet=sheet)(x)] += 1
         for t in list(rtn.keys()):
             rtn[t] = {k:v for k,v in rtn[t].items() if v > 1}
             if not rtn[t]:
                 del(rtn[t])
         return rtn
-    def _sub_tuple(self, table, fields, field_indicies, treat_inf_as_infinity, datemode) :
+    def _get_dv_dt(self, table, field):
+        # reminder - data fields have a default default of zero, primary keys don't get a default default
+        if (table, field) not in self._dv_dt:
+            self._dv_dt[table, field] = (
+                self.tic_dat_factory.default_values.get(table, {}).get(field, ["LIST", "NOT", "POSSIBLE"]),
+                self.tic_dat_factory.data_types.get(table, {}).get(field)
+            )
+        return self._dv_dt[table, field]
+    def _sub_tuple(self, table, fields, field_indicies, treat_inf_as_infinity, sheet) :
         assert set(fields).issubset(field_indicies)
         if self.tic_dat_factory.infinity_io_flag != "N/A" or \
             (table == "parameters" and self.tic_dat_factory.parameters):
             treat_inf_as_infinity = False
         def _read_cell(x, field):
-            # reminder - data fields have a default default of zero, primary keys don't get a default default
-            dv = self.tic_dat_factory.default_values.get(table, {}).get(field, ["LIST", "NOT", "POSSIBLE"])
-            dt = self.tic_dat_factory.data_types.get(table, {}).get(field)
+            dv, dt = self._get_dv_dt(table, field)
             rtn = x[field_indicies[field]]
             if rtn == "" and ((dt and dt.nullable) or (not dt and dv is None)):
                 return None
@@ -268,13 +340,8 @@ class XlsTicFactory(freezable_factory(object, "_isFrozen")) :
                 try_rtn = self.tic_dat_factory._general_read_cell(table, field, None) # None as infinity flagging
                 if utils.numericish(try_rtn):
                     return try_rtn
-            if utils.numericish(rtn) and dt and dt.datetime:
-                rtn = utils.safe_apply(lambda : xlrd.xldate_as_tuple(rtn, datemode))()
-                if rtn is not None:
-                    f = datetime.datetime
-                    if utils.pd:
-                        f = utils.pd.Timestamp
-                    return f(year=rtn[0], month=rtn[1], day=rtn[2], hour=rtn[3], minute=rtn[4], second=rtn[5])
+            if utils.numericish(rtn) and dt and dt.datetime and hasattr(sheet, "xldate_as_tuple_munge"):
+                rtn = sheet.xldate_as_tuple_munge(rtn)
             return self.tic_dat_factory._general_read_cell(table, field, rtn)
         def rtn(x) :
             if len(fields) == 1 :
@@ -316,7 +383,7 @@ class XlsTicFactory(freezable_factory(object, "_isFrozen")) :
                           Needs to end in either ".xls" or ".xlsx"
                           The latter is capable of writing out larger tables,
                           but the former handles infinity seamlessly.
-                          If ".xlsx", then be advised that +/- float("inf") will be replaced
+                          If ".xlsx" then be advised that +/- float("inf") will be replaced
                           with "inf"/"-inf", unless infinity_io_flag is being applied.
 
         :param allow_overwrite: boolean - are we allowed to overwrite an
@@ -386,7 +453,7 @@ class XlsTicFactory(freezable_factory(object, "_isFrozen")) :
         def clean_for_write(t, f, x):
             if self.tic_dat_factory.infinity_io_flag != "N/A" or \
                (t == "parameters" and self.tic_dat_factory.parameters):
-                return self.tic_dat_factory._infinity_flag_write_cell(t, f, x)
+                x = self.tic_dat_factory._infinity_flag_write_cell(t, f, x)
             if x in [float("inf"), -float("inf")] or isinstance(x, datetime.datetime):
                 return str(x)
             return x
@@ -401,7 +468,8 @@ class XlsTicFactory(freezable_factory(object, "_isFrozen")) :
                 for row_ind, (p_key, data) in enumerate(_t.items()) :
                     for field_ind, cell in enumerate( (p_key if containerish(p_key) else (p_key,)) +
                                         tuple(data[_f] for _f in tdf.data_fields.get(t, ()))):
-                        sheet.write(row_ind+1, field_ind, clean_for_write(t, all_flds[field_ind], cell))
+                        write_cell =  clean_for_write(t, all_flds[field_ind], cell)
+                        sheet.write(row_ind+1, field_ind, write_cell)
             else :
                 for row_ind, data in enumerate(_t if containerish(_t) else _t()) :
                     for field_ind, cell in enumerate(tuple(data[_f] for _f in tdf.data_fields[t])) :
